#  This file is a part of the LinkAhead Project.
#
#  Copyright (C) 2024 Henrik tom Wörden (h.tomwoerden@indiscale.com)
#  Copyright (C) 2024 IndiScale GmbH (info@indiscale.com)
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program. If not, see <https://www.gnu.org/licenses/>.

"""
basic integration test of the loan workflow
"""
import json
import re
from functools import partial
from tempfile import NamedTemporaryFile
from unittest.mock import patch

import linkahead as db
from caosadvancedtools.serverside.helper import init_data_model
from linkahead.cached import cached_get_entity_by
from linkahead.utils.get_entity import get_entity_by_id, get_entity_by_name
from linkahead.utils.server_side_scripting import run_server_side_script
from loan.accept_loan_request import accept_loan_request
from loan.accept_return_request import accept_return_request
from loan.box_loan import (BORROWER, BOX, BOX_BORROWED, BOX_RETURNED, COMMENT, CONTENT, DESTINATION,
                           EMAIL, EXHAUST_CONTENTS, EXPECTED_RETURN, F_BOX, F_COMMENT,
                           F_CURRENT_LOCATION, F_DESTINATION, F_EMAIL, F_EXHAUST_CONTENTS,
                           F_EXPECTED_RETURN_DATE, F_FIRST_NAME, F_LAST_NAME, F_LOAN, FIRST_NAME,
                           LAST_NAME, LENT, LOAN, LOAN_ACCEPTED, LOAN_REQUESTED, LOCATION, PERSON,
                           PROPERTIES, RECORD_TYPES, RETURN_ACCEPTED, RETURN_REQUESTED, RETURNED,
                           RETURNLOCATION, assert_date_in_future, assert_key_in_data,
                           insert_or_update_person, main, send_loan_request_mail)
from loan.confirm_loan import confirm_loan
from loan.manual_return import manual_return
from loan.request_loan import issue_loan_request
from loan.request_return import issue_return_request
from pytest import fixture

TESTLOANCOMMENT = 'This is a test'
TESTRETURNCOMMENT = 'This is a return test'
TESTPROP = 'TESTPROP'
TESTBOXNUMBER = '0123 TEST'
TESTBOXNUMBER2 = '0124 TEST'
TESTDESTINATION = 'TEST DEST'
TESTCURRENTDESTINATION = 'TEST CUR DEST'
TESTBORROWEREMAIL = 'test@example.com'
TESTRETURNDATE = "2025-03-04"
TESTFN = "Alice"
TESTLN = "Wunderland"


def save_dict_to_jsonfile(data):
    tmp = NamedTemporaryFile(delete=False, suffix=".json")
    tmp.close()
    with open(tmp.name, "w") as fi:
        json.dump(data, fi)
    return tmp.name


def delete_stuff():
    to_be_deleted = db.Container()
    to_be_deleted.extend(db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'"))
    to_be_deleted.extend(db.execute_query(f"FIND '{TESTDESTINATION}'"))
    to_be_deleted.extend(db.execute_query(f"FIND '{TESTCURRENTDESTINATION}'"))
    to_be_deleted.extend(db.execute_query(f"FIND '{TESTBOXNUMBER}'"))
    to_be_deleted.extend(db.execute_query(f"FIND '{TESTBOXNUMBER2}'"))
    to_be_deleted.extend(db.execute_query(f"FIND person with {EMAIL.name}='{TESTBORROWEREMAIL}'"))
    to_be_deleted.extend(db.execute_query(f"FIND Property '{TESTPROP}'"))
    if len(to_be_deleted) > 0:
        to_be_deleted.delete()


@fixture(autouse=True)
def prepare():
    delete_stuff()
    db.Record(name=TESTBOXNUMBER).add_parent(BOX.name).insert()
    db.Record(name=TESTBOXNUMBER2).add_parent(BOX.name).insert()
    db.Record(name=TESTDESTINATION).add_parent(LOCATION.name).insert()
    db.Record(name=TESTCURRENTDESTINATION).add_parent(LOCATION.name).insert()
    (db.Record().add_parent(PERSON.name).add_property(EMAIL.name, TESTBORROWEREMAIL)
                .add_property(FIRST_NAME.name, TESTFN)
                .add_property(LAST_NAME.name, TESTLN).insert())
    db.Property(name=TESTPROP, datatype=db.TEXT).insert()
    yield
    delete_stuff()


@fixture()
def base_loan_form_data():
    data = {}
    data[F_EXHAUST_CONTENTS] = False
    data[F_EXPECTED_RETURN_DATE] = "2042-02-02"
    data[F_COMMENT] = TESTLOANCOMMENT
    data[F_DESTINATION] = get_entity_by_name(TESTDESTINATION).id
    data[F_EMAIL] = TESTBORROWEREMAIL
    data[F_LAST_NAME] = TESTLN
    data[F_FIRST_NAME] = TESTFN
    return data


@fixture()
def loan_form_data_single(base_loan_form_data):
    base_loan_form_data[F_BOX] = get_entity_by_name(TESTBOXNUMBER).id
    return base_loan_form_data


@fixture()
def loan_form_data_multi(base_loan_form_data):
    base_loan_form_data[F_BOX] = [
        str(get_entity_by_name(TESTBOXNUMBER).id),  # strings are allowed
        get_entity_by_name(TESTBOXNUMBER2).id]
    return base_loan_form_data


def get_return_request_data(loanid):
    data = {}
    data[F_LOAN] = loanid
    # TODO: this is writen in the content field. Change name?
    data[F_COMMENT] = TESTRETURNCOMMENT
    # TODO: this is written in RETURNLOCATION. Change name?
    data[F_CURRENT_LOCATION] = get_entity_by_name(TESTCURRENTDESTINATION).id
    data[F_EXPECTED_RETURN_DATE] = TESTRETURNDATE
    data[F_EMAIL] = TESTBORROWEREMAIL
    data[F_LAST_NAME] = TESTLN
    data[F_FIRST_NAME] = TESTFN
    return data


def test_request_loan(loan_form_data_single):
    data = loan_form_data_single

    # #### request loan ##### #
    response = run_server_side_script(
        "loan_management/request_loan.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr+response.stdout
    assert response.code == 0
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    loanid = loan.id
    assert loan.get_property(f"{BOX.name}").value == data[F_BOX]
    assert loan.get_property(f"{LOAN_REQUESTED.name}").value.startswith("20")
    assert loan.get_property(f"{BORROWER.name}").value == db.execute_query(
        f"FIND {PERSON.name} with {EMAIL.name}='{TESTBORROWEREMAIL}'", unique=True).id
    assert loan.get_property(f"{EXPECTED_RETURN.name}").value == data[F_EXPECTED_RETURN_DATE]
    assert loan.get_property(f"{EXHAUST_CONTENTS.name}").value == data[F_EXHAUST_CONTENTS]
    assert loan.get_property(f"{DESTINATION.name}").value == data[F_DESTINATION]
    assert loan.get_property(f"{LOAN_ACCEPTED.name}") is None
    assert loan.get_property(f"{LENT.name}") is None

    # #### accept loan ##### #
    data = {}
    data[F_LOAN] = loanid
    response = run_server_side_script(
        "loan_management/accept_loan_request.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{LOAN_ACCEPTED.name}").value.startswith("20")
    assert loan.get_property(f"{LENT.name}") is None
    assert loan.get_property(f"{RETURN_REQUESTED.name}") is None

    # #### confirm loan #### #
    data = {}
    data[F_LOAN] = loanid
    response = run_server_side_script(
        "loan_management/confirm_loan.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{LENT.name}").value.startswith("20")

    # #### request return #### #
    response = run_server_side_script(
        "loan_management/request_return.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(get_return_request_data(loan.id))},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{LOAN_ACCEPTED.name}").value.startswith("20")
    assert loan.get_property(f"{CONTENT.name}").value == TESTRETURNCOMMENT
    assert loan.get_property(f"{EXPECTED_RETURN.name}").value == TESTRETURNDATE
    assert loan.get_property(
        f"{RETURNLOCATION.name}").value == get_entity_by_name(TESTCURRENTDESTINATION).id
    assert loan.get_property(f"{RETURN_REQUESTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURN_ACCEPTED.name}") is None
    # TODO test change of borrower

    # accept return
    data = {}
    data[F_LOAN] = loanid
    response = run_server_side_script(
        "loan_management/accept_return_request.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{RETURN_ACCEPTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURNED.name}") is None

    # manual return
    data = {}
    data[F_LOAN] = loanid
    response = run_server_side_script(
        "loan_management/manual_return.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{RETURNED.name}").value.startswith("20")


def test_request_loan_multiple_items(loan_form_data_multi):

    # #### request loan #### #
    data = loan_form_data_multi
    box_ids = loan_form_data_multi[F_BOX]
    box_versionids = [get_entity_by_id(eid=bid).get_versionid() for bid in box_ids]

    response = run_server_side_script(
        "loan_management/request_loan.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    loanid = loan.id
    # We need to cast data[F_BOX] to int, since we may have strings in
    # the form.
    assert loan.get_property(f"{BOX.name}").value == [int(bid) for bid in data[F_BOX]]

    # #### accept loan #### #
    data = {}
    data[F_LOAN] = loan.id
    response = run_server_side_script(
        "loan_management/accept_loan_request.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{LOAN_ACCEPTED.name}").value.startswith("20")

    # #### confirm loan #### #
    data = {}
    data[F_LOAN] = loan.id
    response = run_server_side_script(
        "loan_management/confirm_loan.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = get_entity_by_id(loanid)
    assert loan.get_property(BOX_BORROWED.name).value == box_versionids
    assert loan.get_property(f"{LENT.name}").value.startswith("20")
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{DESTINATION.name}").value
    assert box2.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{DESTINATION.name}").value

    # #### request return #### #
    data = get_return_request_data(loan.id)
    response = run_server_side_script(
        "loan_management/request_return.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr+response.stdout
    assert response.code == 0
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURN_REQUESTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURN_ACCEPTED.name}") is None
    # TODO test change of borrower

    # accept return
    data = {}
    data[F_LOAN] = loan.id
    response = run_server_side_script(
        "loan_management/accept_return_request.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURN_ACCEPTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURNED.name}") is None
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{RETURNLOCATION.name}").value
    assert box2.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{RETURNLOCATION.name}").value

    # manual return
    data = {}
    data[F_LOAN] = loan.id
    response = run_server_side_script(
        "loan_management/manual_return.py",
        # "pos0",
        # option1="val1",
        files={"-p0": save_dict_to_jsonfile(data)},
        **{"auth-token": db.get_connection()._authenticator.auth_token})
    assert response.stderr is None, response.stderr
    assert response.code == 0
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURNED.name}").value.startswith("20")


def test_direct_call(loan_form_data_multi):
    init_data_model(RECORD_TYPES + PROPERTIES)
    data = loan_form_data_multi
    test_prop = cached_get_entity_by(name=TESTPROP)
    box_ids = loan_form_data_multi[F_BOX]
    box_initial_versionids = [get_entity_by_id(eid=bid).get_versionid() for bid in box_ids]
    issue_loan_request(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    loanid = loan.id
    # We need to cast data[F_BOX] to int, since we may have strings in
    # the form.
    assert loan.get_property(f"{BOX.name}").value == [int(bid) for bid in data[F_BOX]]

    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    data = {}
    data[F_LOAN] = loan.id
    accept_loan_request(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{LOAN_ACCEPTED.name}").value.startswith("20")

    # make changes to the boxes after the first request and before loan is confirmed
    for bid in box_ids:
        box = get_entity_by_id(eid=bid)
        box.add_property(id=test_prop.id, value='a')
        box.update()
    box_before_confirm_versionids = [get_entity_by_id(eid=bid).get_versionid() for bid in box_ids]

    data = {}
    data[F_LOAN] = loan.id
    confirm_loan(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(BOX_BORROWED.name).value == box_before_confirm_versionids
    assert loan.get_property(f"{LENT.name}").value.startswith("20")
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{DESTINATION.name}").value
    assert box2.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{DESTINATION.name}").value

    data = get_return_request_data(loan.id)
    issue_return_request(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURN_REQUESTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURN_ACCEPTED.name}") is None

    data = {}
    data[F_LOAN] = loan.id
    accept_return_request(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURN_ACCEPTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURNED.name}") is None
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{RETURNLOCATION.name}").value
    assert box2.get_property(f"{LOCATION.name}").value == loan.get_property(
        f"{RETURNLOCATION.name}").value

    # make changes to the boxes before manual return
    for bid in box_ids:
        box = get_entity_by_id(eid=bid)
        box.add_property(id=test_prop.id, value='a')
        box.update()
    box_before_return_versionids = [get_entity_by_id(eid=bid).get_versionid() for bid in box_ids]

    data = {}
    data[F_LOAN] = loan.id
    manual_return(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    assert loan.get_property(f"{RETURNED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURNED.name}").value.startswith("20")
    assert loan.get_property(BOX_BORROWED.name).value == box_before_confirm_versionids
    assert loan.get_property(BOX_RETURNED.name).value == box_before_return_versionids


def modded_config(config):
    config = dict(config)
    print("changing original config")
    config["Misc"]["entity_loan.no_location_updates"] = "True"
    return config


@patch('linkahead.get_config', new=partial(modded_config, db.get_config()))
def test_no_location_setting(loan_form_data_multi):
    init_data_model(RECORD_TYPES + PROPERTIES)
    data = loan_form_data_multi
    box_ids = loan_form_data_multi[F_BOX]
    box_versionids = [get_entity_by_id(eid=bid).get_versionid() for bid in box_ids]

    issue_loan_request(data)
    loan = db.execute_query(f"FIND loan with {COMMENT.name}='{TESTLOANCOMMENT}'", unique=True)
    loanid = loan.id

    data = {}
    data[F_LOAN] = loanid
    accept_loan_request(data)

    data = {}
    data[F_LOAN] = loanid
    confirm_loan(data)
    loan = get_entity_by_id(loanid)
    assert loan.get_property(BOX_BORROWED.name).value == box_versionids
    assert loan.get_property(f"{LENT.name}").value.startswith("20")
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}") is None
    assert box2.get_property(f"{LOCATION.name}") is None

    data = get_return_request_data(loan.id)
    issue_return_request(data)
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{RETURN_REQUESTED.name}").value.startswith("20")
    assert loan.get_property(f"{RETURN_ACCEPTED.name}") is None

    data = {}
    data[F_LOAN] = loan.id
    accept_return_request(data)
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{RETURN_ACCEPTED.name}").value.startswith("20")
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}") is None
    assert box2.get_property(f"{LOCATION.name}") is None

    data = {}
    data[F_LOAN] = loan.id
    manual_return(data)
    loan = get_entity_by_id(loanid)
    assert loan.get_property(f"{RETURNED.name}").value.startswith("20")
    box1 = get_entity_by_name(TESTBOXNUMBER)
    box2 = get_entity_by_name(TESTBOXNUMBER2)
    assert box1.get_property(f"{LOCATION.name}") is None
    assert box2.get_property(f"{LOCATION.name}") is None
